
#ifndef __PROLOG_KNOWLEDGEBASE_H
    #define __PROLOG_KNOWLEDGEBASE_H

    #include<iostream>
    #include<sstream>
    #include<string>
    #include<vector>
    #include<map>

    namespace Prolog
    {

        // =========================================================================

        class Term
        {
            public:
                virtual ~Term();
                virtual void Portray(std::ostream& os) = 0;
                virtual bool ContainsVariable(std::string* Name) = 0;
                virtual Term* Clone() = 0;
                Term* Clone(int& VariableIndex);
                virtual Term* Clone(int& VariableIndex, std::map<std::string, std::string>& IndexedVariables) = 0;
                static void ReplaceVariables(Term** t, std::string* Name, Term* repl);
        };



        typedef std::vector<Term*> Terms;
        typedef Terms Goal;
        typedef std::vector<std::pair<Term*, Term*> > Unifier;
        typedef std::vector<Goal*> Goals;



        class CompositeTerm : public Term
        {
            public:
                std::string* Head;
                Terms* Children;

                CompositeTerm(std::string* Name, Terms* Children);
                virtual ~CompositeTerm();

                virtual void Portray(std::ostream& os);
                virtual bool ContainsVariable(std::string* Name);
                virtual Term* Clone();
                virtual Term* Clone(int& VariableIndex, std::map<std::string, std::string>& IndexedVariables);
        };

        class Variable : public Term
        {
            public:
                std::string* Name;

                Variable(std::string* Name);
                virtual ~Variable();

                virtual void Portray(std::ostream& os);
                virtual bool ContainsVariable(std::string* Name);
                virtual Term* Clone();
                virtual Term* Clone(int& VariableIndex, std::map<std::string, std::string>& IndexedVariables);
        };


        // =========================================================================

        class Clause
        {
            public:
                Term* Head;
                Terms* Block;

                Clause(Term* Head, Terms* Block);
                virtual ~Clause();

                virtual void Portray(std::ostream& os);
                virtual Clause* Clone(int& VariableIndex);

        };

        typedef std::vector<Clause*> Clauses;

        class KnowledgeBase
        {
            public:
                Clauses* clauses;
                std::map<std::string, Clauses> ClauseRegister;
            public:
                Goals* goals;
                KnowledgeBase();
                virtual ~KnowledgeBase();

                void AddClause(Clause* clause);
                virtual void Portray(std::ostream& os);
                static KnowledgeBase* Parse(std::istream& is);
        };



        Unifier* Unify(Term* t1, Term* t2);
        bool Prove(KnowledgeBase* knowledgebase, Goal* goal, int& VariableCounter);


    } // namespace Prolog

#endif